/*
 *	This is a part of adsbox - ADS-B decoder software
 * 	(c) atty 2011 romulmsg@mail.ru
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "misc.h"

#define MAXDIST 300 // reject too far position
#define EXPIRE	15  // seconds to expire entry
#define CNT_POS 32  // entries in array
typedef struct {
	unsigned int icao;
	float lat0;
	float lon0;
	float lat1;
	float lon1;
	unsigned char last_frame;
	unsigned long timestamp;
} struct_posn;

char cs_tbl[] = { '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K',
		'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y',
		'Z', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ',
		' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', '0', '1', '2', '3', '4', '5',
		'6', '7', '8', '9', ' ', ' ', ' ', ' ', ' ', ' ' };

struct_posn posn[CNT_POS]; // odd and even positions array
unsigned char rownum; // current position row
extern float homelat;
extern float homelon;

float dist(float lat, float lon) {
	float d;

	d = 6371 * acos(sin(lat * M_PI / 180.) * sin(homelat * M_PI / 180.) + cos(
			lat * M_PI / 180.) * cos(homelat * M_PI / 180.) * cos((lon
			- homelon) * M_PI / 180.));
	return (d);
}

float NL(float lat) {
	if (fabs(lat) < 10.47047130)
		return (59);
	else if (fabs(lat) < 14.82817437)
		return (58);
	else if (fabs(lat) < 18.18626357)
		return (57);
	else if (fabs(lat) < 21.02939493)
		return (56);
	else if (fabs(lat) < 23.54504487)
		return (55);
	else if (fabs(lat) < 25.82924707)
		return (54);
	else if (fabs(lat) < 27.93898710)
		return (53);
	else if (fabs(lat) < 29.91135686)
		return (52);
	else if (fabs(lat) < 31.77209708)
		return (51);
	else if (fabs(lat) < 33.53993436)
		return (50);
	else if (fabs(lat) < 35.22899598)
		return (49);
	else if (fabs(lat) < 36.85025108)
		return (48);
	else if (fabs(lat) < 38.41241892)
		return (47);
	else if (fabs(lat) < 39.92256684)
		return (46);
	else if (fabs(lat) < 41.38651832)
		return (45);
	else if (fabs(lat) < 42.80914012)
		return (44);
	else if (fabs(lat) < 44.19454951)
		return (43);
	else if (fabs(lat) < 45.54626723)
		return (42);
	else if (fabs(lat) < 46.86733252)
		return (41);
	else if (fabs(lat) < 48.16039128)
		return (40);
	else if (fabs(lat) < 49.42776439)
		return (39);
	else if (fabs(lat) < 50.67150166)
		return (38);
	else if (fabs(lat) < 51.89342469)
		return (37);
	else if (fabs(lat) < 53.09516153)
		return (36);
	else if (fabs(lat) < 54.27817472)
		return (35);
	else if (fabs(lat) < 55.44378444)
		return (34);
	else if (fabs(lat) < 56.59318756)
		return (33);
	else if (fabs(lat) < 57.72747354)
		return (31);
	else if (fabs(lat) < 58.84763776)
		return (30);
	else if (fabs(lat) < 59.95459277)
		return (30);
	else if (fabs(lat) < 61.04917774)
		return (29);
	else if (fabs(lat) < 62.13216659)
		return (28);
	else if (fabs(lat) < 63.20427479)
		return (27);
	else if (fabs(lat) < 64.26616523)
		return (26);
	else if (fabs(lat) < 65.31845310)
		return (25);
	else if (fabs(lat) < 66.36171008)
		return (24);
	else if (fabs(lat) < 67.39646774)
		return (23);
	else if (fabs(lat) < 68.42322022)
		return (22);
	else if (fabs(lat) < 69.44242631)
		return (21);
	else if (fabs(lat) < 70.45451075)
		return (20);
	else if (fabs(lat) < 71.45986473)
		return (19);
	else if (fabs(lat) < 72.45884545)
		return (18);
	else if (fabs(lat) < 73.45177442)
		return (17);
	else if (fabs(lat) < 74.43893416)
		return (16);
	else if (fabs(lat) < 75.42056257)
		return (15);
	else if (fabs(lat) < 76.39684391)
		return (14);
	else if (fabs(lat) < 77.36789461)
		return (13);
	else if (fabs(lat) < 78.33374083)
		return (12);
	else if (fabs(lat) < 79.29428225)
		return (11);
	else if (fabs(lat) < 80.24923213)
		return (10);
	else if (fabs(lat) < 81.19801349)
		return (9);
	else if (fabs(lat) < 82.13956981)
		return (8);
	else if (fabs(lat) < 83.07199445)
		return (7);
	else if (fabs(lat) < 83.99173563)
		return (6);
	else if (fabs(lat) < 84.89166191)
		return (5);
	else if (fabs(lat) < 85.75541621)
		return (4);
	else if (fabs(lat) < 86.53536998)
		return (3);
	else if (fabs(lat) < 87.00000000)
		return (2);
	else
		return (1);
}

unsigned char insert_posn(unsigned char f, unsigned int icao, int lat, int lon) {
	unsigned char i;
	time_t t;

	t = time(NULL);

	for (i = 0; i < CNT_POS; i++) {
		if (posn[i].icao == icao) { // icao code found
			if (f == 0) {
				posn[i].lat0 = lat;
				posn[i].lon0 = lon;
				if ((t - posn[i].timestamp) > EXPIRE) {
					posn[i].lat1 = 0; // those coords are expired
					posn[i].lon1 = 0;
				}
			} else {
				posn[i].lat1 = lat;
				posn[i].lon1 = lon;
				if ((t - posn[i].timestamp) > EXPIRE) {
					posn[i].lat0 = 0;
					posn[i].lon0 = 0;
				}
			}
			posn[i].last_frame = f;
			posn[i].timestamp = t;
			return (i);
		}
	}

	rownum = (rownum + 1) % CNT_POS; // icao code does not found
	posn[rownum].lat0 = 0;
	posn[rownum].lon0 = 0;
	posn[rownum].lat1 = 0;
	posn[rownum].lon1 = 0;

	if (f == 0) {
		posn[rownum].lat0 = lat;
		posn[rownum].lon0 = lon;
	} else {
		posn[rownum].lat1 = lat;
		posn[rownum].lon1 = lon;
	}
	posn[rownum].icao = icao;
	posn[rownum].last_frame = f;
	posn[rownum].timestamp = t;
	return (rownum);
}

float modulo(float x, float y) {
	float r;
	r = x - y * floor(x / y);
	return (r);
}

int crc(char * buf) {
	int poly;
	int data, data1, data2;
	int i;

	poly = 0xfffa0480;
	sscanf(buf, "*%08x%08x%08x", &data, &data1, &data2);
	data2 = data2 & 0xffffff00;
	for (i = 0; i < 88; i++) {
		if ((data & 0x80000000) != 0)
			data = data ^ poly;
		data = data << 1;
		if ((data1 & 0x80000000) != 0)
			data = data | 1;
		data1 = data1 << 1;
		if ((data2 & 0x80000000) != 0)
			data1 = data1 | 1;
		data2 = data2 << 1;
	}
	return (data);
}

int decode_message(char * buf, char * msg) {
	unsigned char rawdata[14];
	unsigned char df;
	unsigned char tc;
	char callsign[9];
	unsigned int icao;
	unsigned int cs;
	unsigned char f;
	int lat, lon;
	int alt;
	unsigned char i;
	int j, ni, m;
	float rlat0, rlat1;
	float dlon;
	float rlat, rlon;
	float d;
	int c;
	unsigned char subtype;
	double ew_spd, ns_spd;
	double gnd_spd, trk;
	int vr;
	unsigned char ew_dir, ns_dir;

	sscanf(
			buf,
			"*%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx",
			&rawdata[0], &rawdata[1], &rawdata[2], &rawdata[3], &rawdata[4],
			&rawdata[5], &rawdata[6], &rawdata[7], &rawdata[8], &rawdata[9],
			&rawdata[10], &rawdata[11], &rawdata[12], &rawdata[13]);

	df = (rawdata[0] >> 3) & 0x1F;
	DEBUG("DF:%u\n", df);

	if (df == 17) {
		c = crc(buf);
		if (c != ((rawdata[11] << 24) | (rawdata[12] << 16) | rawdata[13] << 8)) {
			DEBUG("Bad CRC !\n", c);
			return (0);
		}
		icao = (rawdata[1] << 16) | (rawdata[2] << 8) | rawdata[3];
		tc = (rawdata[4] >> 3) & 0x1f;
		DEBUG("ICAO:%06x TC:%u\n", icao, tc);

		if (tc == 4) { // aircraft ID
			bzero(&callsign, 9);
			cs = (rawdata[5] << 16) | (rawdata[6] << 8) | (rawdata[7]);
			callsign[0] = cs_tbl[(cs >> 18) & 0x3f];
			callsign[1] = cs_tbl[(cs >> 12) & 0x3f];
			callsign[2] = cs_tbl[(cs >> 6) & 0x3f];
			callsign[3] = cs_tbl[cs & 0x3f];
			cs = (rawdata[8] << 16) | (rawdata[9] << 8) | rawdata[10];
			callsign[4] = cs_tbl[(cs >> 18) & 0x3f];
			callsign[5] = cs_tbl[(cs >> 12) & 0x3f];
			callsign[6] = cs_tbl[(cs >> 6) & 0x3f];
			callsign[7] = cs_tbl[cs & 0x3f];

			DEBUG("Callsign %s\n", callsign);
			sprintf(msg, "MSG,1,,,%06X,,,,,,%s,,,,,,,,0,0,0,0\n\0", icao,
					callsign);
		}

		if ((tc > 8) && (tc < 19)) { // aircraft air position
			f = (rawdata[6] >> 2) & 0x01;
			lat = ((rawdata[8] >> 1) & 0x7f) | (rawdata[7] << 7) | ((rawdata[6]
					& 0x03) << 15);
			lon = rawdata[10] | (rawdata[9] << 8) | ((rawdata[8] & 0x01) << 16);
			alt = 25 * (((rawdata[6] >> 4) & 0x0f)
					| (((rawdata[5] >> 1) & 0x7f) << 4)) - 1000;
			i = insert_posn(f, icao, lat, lon);

			if ((posn[i].lat0 > 0) && (posn[i].lat1 > 0)) {
				j = floor((59. * posn[i].lat0 - 60. * posn[i].lat1) / 131072.
						+ 0.5);

				rlat0 = 6. * (modulo(j, 60) + posn[i].lat0 / 131072.);
				rlat1 = 6.101694915254237288 * (modulo(j, 59) + posn[i].lat1
						/ 131072.);

				rlat0 = (rlat0 < 270 ? rlat0 : rlat0 - 360.);
				rlat1 = (rlat1 < 270 ? rlat1 : rlat1 - 360.);

				if (NL(rlat0) == NL(rlat1)) {
					if (f == 0) {
						ni = (NL(rlat0) > 1 ? NL(rlat0) : 1);
						dlon = 360. / ni;
						m = floor((posn[i].lon0 * (NL(rlat0) - 1)
								- posn[i].lon1 * NL(rlat0)) / 131072. + 0.5);
						rlat = rlat0; //(rlat0 < 270 ? rlat0 : rlat0 - 360.);
						rlon = dlon * (modulo(m, ni) + posn[i].lon0 / 131072.);
					} else {
						ni = ((NL(rlat1) - 1) > 1 ? (NL(rlat1) - 1) : 1);
						m = floor((posn[i].lon0 * (NL(rlat1) - 1)
								- posn[i].lon1 * NL(rlat1)) / 131072. + 0.5);
						dlon = 360. / ni;
						rlat = rlat1; //(rlat1 < 270 ? rlat1 : rlat1 - 360.);
						rlon = dlon * (modulo(m, ni) + posn[i].lon1 / 131072.);
					}
					if ((homelat != 0) && (homelon != 0)) {
						d = dist(rlat, rlon);
						if (d > MAXDIST)
							return (0); //reject too far position
						DEBUG("Distance %f ", d);
					}
					DEBUG("lat %1.5f lon %1.5f\n", rlat, rlon);
					sprintf(
							msg,
							"MSG,3,,,%06X,,,,,,,%d,,,%1.5f,%1.5f,,,0,0,0,0\n\0",
							icao, alt, rlat, rlon);
				}
			} //else
			//  DEBUG("No odd and even both.\n");
		}
		if (tc == 19) { // air speed
			subtype = rawdata[4] & 0x07;
			if (subtype == 1) { // ground non supersonic speed
				ew_spd = rawdata[6] | ((rawdata[5] & 0x03) << 8); // -1 TODO
				ns_spd = ((rawdata[8] >> 5) & 0x07)
						| ((rawdata[7] & 0x7f) << 3);
				if ((ew_spd == 0) && (ns_spd == 0))
					return (0);

				ew_dir = (rawdata[5] >> 2) & 0x01;
				ns_dir = (rawdata[7] >> 7) & 0x01;
				gnd_spd = floor(sqrt(ew_spd * ew_spd + ns_spd * ns_spd));
				if ((!ew_dir) && (!ns_dir))
					trk = (ew_spd == 0 ? 0 : 90 - 180. / M_PI * atan(ns_spd
							/ ew_spd));
				if ((!ew_dir) && (ns_dir))
					trk = (ew_spd == 0 ? 180 : 90 + 180. / M_PI * atan(ns_spd
							/ ew_spd));
				if ((ew_dir) && (ns_dir))
					trk = (ew_spd == 0 ? 180 : 270 - 180. / M_PI * atan(ns_spd
							/ ew_spd));
				if ((ew_dir) && (!ns_dir))
					trk = (ew_spd == 0 ? 0 : 270 + 180. / M_PI * atan(ns_spd
							/ ew_spd));
				vr = ((rawdata[8] & 0x08) == 0 ? 1 : -1) * ((((rawdata[9] >> 2)
						& 0x3f) | (rawdata[8] & 0x07) << 6) - 1) * 64;
				sprintf(msg,
						"MSG,4,,,%06X,,,,,,,,%1.1f,%1.1f,,,%i,,0,0,0,0\n\0",
						icao, gnd_spd, trk, vr);
			}
		}
	}
	return (strlen(msg));
}
