/*
 *	adsbox - ADS-B decoder software
 *	(c) atty 2011 romulmsg@mail.ru
 *
 */

#include <stdio.h>
#include <stdlib.h> 
#include <fcntl.h>
#include <termios.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <pthread.h>
#include <time.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include "misc.h"

#define BUFLEN 256
#define PORT 30003
#define BAUDRATE B115200
#define DEVICE "/dev/ttyS1"

pthread_mutex_t mux;
pthread_cond_t cond_var;
char gdata[BUFLEN];
int dmode = 0;
float homelat = 0;
float homelon = 0;

void *decoder(void *arg) { /* decoder thread */
	int i;
	i = 0;
	int fd = (int) arg;
	unsigned char idx;
	unsigned char rawbuf[BUFLEN];
	unsigned char msg[128];

	while (1) {
		bzero(rawbuf, BUFLEN);
		read(fd, rawbuf, BUFLEN);
		if (strlen(rawbuf) == 31) {
			if (decode_message(rawbuf, msg)) {
				pthread_mutex_lock(&mux);
				bzero(gdata, BUFLEN);
				memcpy(gdata, msg, strlen(msg));
				pthread_cond_broadcast(&cond_var);
				pthread_mutex_unlock(&mux);
				bzero(msg, 128);
			}
		}
	}
	pthread_exit(0);
}

void *servlet(void *arg) /* process client connection thread */
{
	char buf[2];
	char cdata[BUFLEN];
	struct timespec to;

	int s = (int) arg;

	while (recv(s, buf, 1, MSG_DONTWAIT)) {
		pthread_mutex_lock(&mux);
		to.tv_sec = time(NULL) + 1;
		to.tv_nsec = 0;
		pthread_cond_timedwait(&cond_var, &mux, &to);
		if (strncmp(cdata, gdata, strlen(gdata))) {
			bzero(cdata, BUFLEN);
			memcpy(cdata, gdata, strlen(gdata));
			pthread_mutex_unlock(&mux);
			send(s, cdata, strlen(cdata), 0);
		}
		pthread_mutex_unlock(&mux);
	}
	close(s);
	pthread_exit(0);
}

int main(int ac, char *av[]) {
	extern char * optarg;
	char s[16] = DEVICE;
	struct sockaddr_in addr;
	int fd, sd, sd1, en = 1;
	struct termios oldtio, newtio;
	int pid, sid;

	pthread_mutex_init(&mux, NULL);

	printf("adsbox ADS-B decoder version %s.\n", VERSION);
	int c;
	while ((c = getopt_long(ac, av, "ds:l:g:")) != -1) {
		switch (c) {
		case 'd':
			dmode = 1;
			break;
		case 's':
			strncpy(s, optarg, 16);
			break;
		case 'l':
			homelat = atof(optarg);
			break;
		case 'g':
			homelon = atof(optarg);
			break;
		case '?':
			printf(
					"%s [options]\n-s input serial device\n-d be a daemon\n-l our latitude\n-g our longitude\n",
					av[0]);
			exit(0);
		default:
			break;
		}
	}
	printf("Using %s as input serial device.\n", s);
	if ((homelat != 0) && (homelon != 0))
		printf("Your lat %2.5f lon %3.5f\n", homelat, homelon);

	/* COMPORT STUFF
	 */
	fd = open(s, O_RDONLY | O_NOCTTY);
	if (fd < 0)
		panic("open device");

	tcgetattr(fd, &oldtio); // TODO restore old termio

	bzero(&newtio, sizeof(newtio));
	newtio.c_cflag &= ~PARENB;
	newtio.c_cflag &= ~CSTOPB;
	newtio.c_cflag &= ~CSIZE;
	newtio.c_cflag |= BAUDRATE | CS8 | CLOCAL | CREAD;
	newtio.c_iflag = IGNPAR | ICRNL; // ignore parity error
	newtio.c_oflag = 0;

	newtio.c_lflag = ICANON;

	newtio.c_cc[VTIME] = 0;
	newtio.c_cc[VMIN] = 1;
	newtio.c_cc[VEOL] = 0;
	newtio.c_cc[VEOL2] = 0;

	tcflush(fd, TCIFLUSH);
	tcsetattr(fd, TCSANOW, &newtio);

	/* COMPORT STUFF
	 */

	//char msg[256];
	//decode_message("8D4000FB5875444148D9A70B023A", msg);
	//exit(0);

	/*--- create socket ---*/
	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (sd < 0)
		panic("socket");
	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &en, sizeof(en)) != 0)
		panic("setsockopt");

	/*--- bind  ---*/
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(PORT);
	addr.sin_addr.s_addr = htonl(INADDR_ANY); /* any interface */
	if (bind(sd, (struct sockaddr*) &addr, sizeof(addr)) != 0)
		panic("bind");

	if (dmode) { /* daemonize */
		pid = fork();
		if (pid < 0)
			panic("fork()");

		if (pid > 0)
			exit(0);

		umask(0);

		sid = setsid();

		if (sid < 0)
			panic("setsid()");

		chdir("/");

		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
	}

	/*--- make into listener with 10 slots ---*/
	if (listen(sd, 10) != 0)
		panic("listen")

	/*--- begin waiting for connections ---*/
	else {
		pthread_t child;

		pthread_create(&child, NULL, decoder, (void *) fd); /* start decoder thread */

		while (1) /* process incoming clients */
		{
			sd1 = accept(sd, 0, 0); /* accept connection */
			pthread_create(&child, NULL, servlet, (void *) sd1); /* start connection thread */
		}
	}
	exit(0);
}
